//---------------------------------------------------------------------------

#pragma hdrstop

#include "avroo.h"
//---------------------------------------------------------------------------

#include <stdio.h>
#include <stdint.h>
#include <string.h>
//#include <conio.h>
//#include <Windows.h>
#include <errno.h>




#include "serial.h"
#include "print.h"

#define CRC_SEED	0x0000

unsigned int g_pagesize_flash	= 512;
unsigned int g_pagesize_eeprom	= 512;


uint16_t crc16_update(uint16_t crc, uint8_t a)
{
	int i;

	crc ^= a;
	for (i = 0; i < 8; ++i)
	{
		if (crc & 1)
			crc = (crc >> 1) ^ 0xA001;
		else
			crc = (crc >> 1);
	}

	return crc;
}

uint16_t crc16(const char* buf, int len, uint16_t crc = CRC_SEED)
{
	for (int i=0; i<len; i++)
	{
		crc = crc16_update(crc, buf[i]);
	}

	return crc;
}

bool binfiletomem(const char* filename, char* buf, int* size)
{
	FILE* file = fopen(filename, "r");
	
	int _size = *size;
	if (file)
	{
		*size = fread(buf, sizeof(char), *size, file);
		fclose(file);
		file = NULL;
	}

	return (0 < *size);
}

bool intelhex2mem(const char* filename, char* image, int* size)
{
	bool ok = true;

	FILE *file = fopen(filename, "r" );
	if (!file)
	{
		eprintf("Error: opening file %s for reading failed: %s \r\n", filename, GetLastErrorText());
		return false;
	}

	memset(image, 0xaa, *size);

	int n = 0;

	char buf[256];

	while (fgets(buf, sizeof(buf), file) != NULL)
	{
		char* str = buf;
		if (str[0] != ':') break;
		str++;

		unsigned int sum = 0x00;

		int len;
		sscanf(str, "%02x", &len);
		str += LITERAL_STRLEN("ll");
		sum += len;

		int addr;
		sscanf(str, "%04x", &addr);
		str += LITERAL_STRLEN("aaaa");
		sum += (addr >> 8) + (addr & 0xff);

		int type;
		sscanf(str, "%02x", &type);
		str += LITERAL_STRLEN("tt");
		sum += type;

		uint8_t q[64];
		for (int j=0; j<len; j++)
		{
			unsigned int d;
			sscanf(str, "%02x", &d);
			str += LITERAL_STRLEN("dd");
			sum += d;
			q[j] = d;

			if (type==0x00)
			{
				if (n < *size)
				{
					image[n++] = d;
				}
				else
				{
					eprintf("Error: not enough buffer space for intel-hexfile.\r\n");
					return false;
                }
			}
		}

		unsigned int chk;
		sscanf(str, "%02x", &chk);
		str += LITERAL_STRLEN("cc");

		if (chk != (((sum^0xff)+0x01)&0xff))
		{
			eprintf("Error: checksum error in intel-hexfile %s!\r\n", filename);
			ok = false;
		}
	}

	fclose(file);
	file = NULL;
	
	*size  = n;

	return ok;
}

bool filetomem(const char* filename, char* buf, int* size)
{
	const char* fileext = filename + strlen(filename) - LLEN(".xyz");

	if (STREQU(fileext, ".hex") || STREQU(fileext, ".eep"))
	{
		//printf("Assuming IntelHex-File.\r\n", filename);
		return intelhex2mem(filename, buf, size);
	}
	else if STREQU(fileext, ".bin")
	{
		//printf("Assuming Binary-File.\r\n", filename);
		return binfiletomem(filename, buf, size);
	}
	else
	{
		eprintf("Error: cant determine file type (filename extension) of %s!\r\n", filename);
		return false;
	}
}

bool avroo_recvreply(int timeout=100, bool verbose=true)
{
	char c = 0xff;
	int r = serial_gets(&c, sizeof(c), timeout);

	if (r < sizeof(char))
	{
		if (verbose) eprintf("Error: Revceive Timeout.\n");
		return false;
	}

	switch (c)
	{
		case 0x30:
		{
			if (verbose) printf_green("Reply OK.\n");
			return true;
		}	
		case 0xc0:
		{
			if (verbose) eprintf("Error: Verification error.\n");
			return false;
		}
		case 0xc1:
		{
			if (verbose) eprintf("Error: Unknown command error.\n");
			return false;
		}
		case 0xc2:
		{
			if (verbose) eprintf("Error: CRC error.\n");
			return false;
		}
		case 0xc3:
		{
			if (verbose) eprintf("Error: Boundary error.\n");
			return false;
		}
		case 0xc4:
		{
			if (verbose) eprintf("Error: Decryption error.\n");
			return false;
		}
		case 0xc5:
		{
			if (verbose) eprintf("Error: Programming error.\n");
			return false;
		}
		case 0xc6:
		{
			if (verbose) eprintf("Error: Wrong version error.\n");
			return false;
		}
		default:
		{
			if (verbose) eprintf("Error: Unknown error: Code 0x%x\n", c & 0xff);
			return false;
		}
	}
}

int avroo_sendbufcrc(const char* buf, int len, uint16_t crc = CRC_SEED)
{
	int res = 0;
	crc = crc16(buf, len, crc);
	res += serial_puts(buf, len);
	res += serial_putc(LOBYTE(crc));
	res += serial_putc(HIBYTE(crc));
	return res;
}

void avroo_sendchars(char c, int n)
{
	while (n--)
		serial_putc(c);
}

int avroo_readbootinfo(int timeout)
{
	int te = GetTickCount_ms() + timeout;

	char rxbuf[64];

	for (int i=0; i<sizeof(rxbuf); i++)
	{
		if (GetTickCount_ms() > te)
		{
			return 0;
		}

		char c;
		int r = serial_gets(&c, sizeof(c), timeout);

		if (r > 0)
		{
			rxbuf[i] = c;
		}

		if IN_RANGE(0x30, c, 0x3f) // msste den wert haben: 97 05 06 04 30
		{
			printhex(rxbuf, i+1);
			// todo: den bootstring analysieren:
			// chip-type, ee-pagesize, flash-pagesize anpassen, bootloader capabilities, ...
			return i + 1;
		}
	}
}

bool avroo_catch(const char* trigger, int triggerlen, int timeout)
{
	bool infinite = (timeout == 0);

	int te = GetTickCount_ms() + timeout;

	printf("Catching Bootloader ... ");

	while (infinite || (GetTickCount_ms() < te))
	{
		serial_flush();
		avroo_sendchars('\0', 32);
		serial_putc('\x0d');
		avroo_sendbufcrc(trigger, triggerlen);

		int rxcnt = avroo_readbootinfo(5);

		if (rxcnt > 0)
		{
			printf_green(" success!\r\n");
			return true;
		}

		sleep_ms(10);
	}

	printf_red(" Timeout!\r\n");
	return false;
}

bool avroo_keepalive(void)
{
	const char buf[] = { 0xfd, 0x00 };
	avroo_sendbufcrc(buf, sizeof(buf));
	return avroo_recvreply(15, false);
}

int avroo_startfirmware(void)
{
	const char buf[] = { 0x00, 0x01 };

	printf("Exiting Bootloader, Starting Firmware. \r\n");
	avroo_sendbufcrc(buf, sizeof(buf));

	//avroo_recvreply();

	return true;
}

int avroo_setaddr(uint32_t addr)
{
	const char buf[] = { 0xff, BYTE2_32(addr), BYTE1_32(addr), BYTE0_32(addr) };
	//printf("Setting Address Pointer to 0x%06lx ... ", 0x00ffffff & addr);
	avroo_sendbufcrc(buf, sizeof(buf));
	return avroo_recvreply(100, false);
}

bool avr_fillpagebuf(const char* buf, int len)
{
	//printf("Sending %5u Bytes to Buffer ... ", len);
	const char hdr[] = { 0xfe, 0x00, BYTE1_32(len), BYTE0_32(len) };
	avroo_sendbufcrc(hdr, sizeof(hdr));
	avroo_sendbufcrc(buf, len);
	return avroo_recvreply(100, false);
}

bool avroo_writebuf2ee(void)
{
	const char buf[] = { 0x05, 0x01 };
	//printf("Writing Buffer to EEPROM ...      ");
	avroo_sendbufcrc(buf, sizeof(buf));
	return avroo_recvreply(5000, false);
}

bool avroo_writebuf2flash(void)
{
	const char buf[] = { 0x01, 0x01 };
	//printf("Writing Buffer to Flash ...       ");
	avroo_sendbufcrc(buf, sizeof(buf));
	return avroo_recvreply(5000, false);
}

enum { MEMTYPE_FLASH, MEMTYPE_EEPROM };

bool avroo_upload_file(const char* filename, int memtype)
{
	if (!filename) return true;
	
	static char image[256 * 1024];
	memset(image, 0xff, sizeof(image));
	int size = sizeof(image);
	
	int page_size = 512;
	bool res = true;

	if (filetomem(filename, image, &size))
	{
		switch (memtype)
		{
			case MEMTYPE_FLASH:
			{
				printf("Uploading File %s (%u Bytes) to Flash... \r\n", filename, size);
				page_size = g_pagesize_flash;
				break;
			}
			case MEMTYPE_EEPROM:
			{
				printf("Uploading File %s (%u Bytes) to EEPROM... \r\n", filename, size);
				page_size = g_pagesize_eeprom;
				break;
			}
		}

		avroo_setaddr(0x000000);

		for (int i=0; i<size; i += page_size)
		{
			//int chunk = (i + page_size > size) ? (size - i) : page_size;
			int chunk = page_size;

			while (!avr_fillpagebuf(image + i, chunk))
			{
				sleep_ms(10);
				printf_yellow("retrying!   \r");
			}

			switch (memtype)
			{
				case MEMTYPE_FLASH:
				{
					res = avroo_writebuf2flash() && res;
					break;
				}
				case MEMTYPE_EEPROM:
				{
					res = avroo_writebuf2ee() && res;
					break;
				}
			}
			
			printf("%c %5u Bytes (%3u %%)\r", wheel(), i, (100*i)/size);
		}

		if (res)
		{
			printf_green("Upload succeded.     \r\n");
		}
		else
		{
        	eprintf("Error: Upload failed!");
		}
	}
	else
	{
		res = false;
	}

	return res;
}

bool avroo_upload(
	const char* portname, const char* cfgstr,
	const char* trigger, int triggerlen,
	const char* flashfile,
	const char* eepfile
)
{
	//triggerlen = triggerlen ? triggerlen : strlen(trigger);

	if (!serial_open(portname, cfgstr))
	{
		return false;
	}

	bool res = avroo_catch(trigger, triggerlen, BOOTLOADER_CATCH_TIMEOUT_MS);

	if (res)
	{
		res = avroo_upload_file(flashfile, MEMTYPE_FLASH)	&& res;
		res = avroo_upload_file(eepfile,   MEMTYPE_EEPROM) 	&& res;

		printf("Hit any Key to exit Bootloader mode...\r\n");
		while (!kbhit())
		{
			printf("%c \r", wheel());
			bool ok = avroo_keepalive();
			sleep_ms(200);
		}

		avroo_startfirmware();

		char rxbuf[64];
		int rxcnt = serial_readline(rxbuf, sizeof(rxbuf) - 1, 500, '>');
		if (rxcnt) { rxbuf[rxcnt] = '\0'; printf("Rx: %s\r\n", rxbuf); }

	}
	else
	{
		eprintf("Error: Bootloader not responding!\r\n");
	}

	serial_close();

	return res;
}